package ch.polgrabia.demos.client;

import ch.polgrabia.demos.utils.TopicUtil;
import io.vertx.core.AbstractVerticle;
import io.vertx.core.AsyncResult;
import io.vertx.core.buffer.Buffer;
import io.vertx.core.http.*;
import io.vertx.core.http.impl.WebSocketClientImpl;
import io.vertx.core.impl.CloseFuture;
import io.vertx.core.impl.VertxInternal;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.concurrent.CompletableFuture;

public class WebsocketClientParticle extends AbstractVerticle {
    private static final Logger logger = LoggerFactory.getLogger(WebsocketClientParticle.class);
    private final String hostname;
    private final int port;
    // path maps to channel
    private final String path;
    private final String channel;
    private HttpClient client;
    private boolean isShuttingDownActivated = false;
    private final Object runningMonitor = new Object();
    private HttpServer server;
    private WebSocket webSocket;

    public WebsocketClientParticle(String hostname, int port, String path) {
        this.hostname = hostname;
        this.port = port;
        this.path = path;
        this.channel = TopicUtil.convertTopicName(path);
    }

    @Override
    public void start() {
        this.client = getVertx()
                .createHttpClient();
        HttpServerOptions httpServerOptions = new HttpServerOptions();
        this.server = getVertx().createHttpServer(httpServerOptions);
        this.server.requestHandler(this::handleCommandRequests);
        this.server.listen(8123, "localhost", this::handleListenCommand);
        logger.info("Starting websocket client command server");
        new WebSocketClientImpl((VertxInternal) getVertx(), new HttpClientOptions(), new CloseFuture())
                .webSocket(port, hostname, path, this::handleMessage);
    }

    private void handleListenCommand(AsyncResult<HttpServer> httpServerAsyncResult) {
        if (httpServerAsyncResult.failed()) {
            logger.warn("Couldn't initialized http command server");
            getVertx().close();
            System.exit(1);
        }

        logger.info("Successfully started websocket client command server");
    }

    private void handleCommandRequests(HttpServerRequest httpServerRequest) {
        logger.info("Received command request on 8123");
        HttpServerResponse response = httpServerRequest.response();
        String qParam = httpServerRequest.getParam("q");
        if (qParam == null || qParam.isBlank()) {
            response.setStatusCode(400);
            response.end("bad request");
            return;
        }
        logger.info("Sending binary message: {}", qParam);
        webSocket.writeBinaryMessage(Buffer.buffer(qParam));
        response.setStatusCode(202);
        response.end("Ok");
    }

    private void handleMessage(AsyncResult<WebSocket> webSocketAsyncResult) {
        logger.info("Handling websocket");
        if (!webSocketAsyncResult.succeeded()) {
            logger.warn("Websocket handler connection failed", webSocketAsyncResult.cause());
            try {
                stop();
            } catch (Exception e) {
                logger.error("Failed to stop particle", e);
            }
            return;
        }
        this.webSocket = webSocketAsyncResult.result();
        vertx.setPeriodic(100, new ClientPeriodicStdinScanner(
                webSocket,
                System.in
        ));

        webSocket
                .binaryMessageHandler(msg -> logger.info("[{}] Got message: {}", channel, msg))
                .closeHandler(unused -> {
                    try {
                        stop();
                    } catch (Exception e) {
                        logger.error("Got an error while stopping", e);
                    }
                });

    }

    @Override
    public void stop() {
        isShuttingDownActivated = true;
        client.close();
        server.close();
        synchronized (runningMonitor) {
            runningMonitor.notifyAll();
        }
    }

    public CompletableFuture<Boolean> waitingUntilFinished() {
        return CompletableFuture.supplyAsync(() -> {
            while (!isShuttingDownActivated) {
                synchronized (runningMonitor) {
                    try {
                        runningMonitor.wait();
                        return true;
                    } catch (InterruptedException e) {
                        logger.warn("Got interrupted", e);
                    }
                }
            }
            return false;
        });
    }

}
